unit SimpleThreadClass;

{$IFDEF VER90}{$DEFINE Delphi_bis_5}{$ENDIF}//Delphi 2
{$IFDEF VER100}{$DEFINE Delphi_bis_5}{$ENDIF}//Delphi 3
{$IFDEF VER120}{$DEFINE Delphi_bis_5}{$ENDIF}//Delphi 4
{$IFDEF VER130}{$DEFINE Delphi_bis_5}{$ENDIF}//Delphi 5

{$IFNDEF Delphi_bis_5}
{$WARN SYMBOL_DEPRECATED OFF}
{$ENDIF}

{$R-}

interface

uses SysUtils,Windows;

type TNotifyEvent=procedure(Sender:TObject) of object;

  EThread=class(Exception);
  TThreadMethod=procedure of object;
  TThreadPriority=(tpIdle,tpLowest,tpLower,tpNormal,tpHigher,tpHighest,tpTimeCritical);

  TSimpleThread=class
  private
    FHandle:THandle;
    FThreadID:THandle;
    FTerminated:Boolean;
    FSuspended:Boolean;
    FMainThreadWaiting:Boolean;
    FFreeOnTerminate:Boolean;
    FFinished:Boolean;
    FReturnValue:Integer;
    FOnTerminate:TNotifyEvent;
    FMethod:TThreadMethod;
    FSynchronizeException:TObject;
    procedure CallOnTerminate;
    function GetPriority:TThreadPriority;
    procedure SetPriority(Value:TThreadPriority);
    procedure SetSuspended(Value:Boolean);
  protected
    procedure DoTerminate;virtual;
    procedure Execute;virtual;abstract;
    procedure Synchronize(Method:TThreadMethod);
    property ReturnValue:Integer read FReturnValue write FReturnValue;
    property Terminated:Boolean read FTerminated;
  public
    constructor Create(CreateSuspended:Boolean);
    destructor Destroy;override;
    procedure Resume;
    procedure Suspend;
    procedure Terminate;
    function WaitFor:dword;//Integer;
    property FreeOnTerminate:Boolean read FFreeOnTerminate write FFreeOnTerminate;
    property Handle:THandle read FHandle;
    property Priority:TThreadPriority read GetPriority write SetPriority;
    property Suspended:Boolean read FSuspended write SetSuspended;
    property ThreadID:THandle read FThreadID;
    property OnTerminate:TNotifyEvent read FOnTerminate write FOnTerminate;
  end;

var MainThreadID:THandle;

implementation

const CM_EXECPROC=$8FFF;

type
  PRaiseFrame=^TRaiseFrame;
  TRaiseFrame=record
    NextRaise:PRaiseFrame;
    ExceptAddr:Pointer;
    ExceptObject:TObject;
    ExceptionRecord:PExceptionRecord
  end;

var
  ThreadWindow:HWND;
  ThreadCount:Integer;

function ThreadWndProc(Window:HWND;Message,wParam,lParam:Longint):Longint;stdcall;
begin
{case Message of
  CM_EXECPROC:}
if Message=CM_EXECPROC then with TSimpleThread(lParam) do
  begin
  Result:=0;
    try
      FSynchronizeException:=nil;
      FMethod
    except
      if RaiseList<>nil then
        begin
        FSynchronizeException:=PRaiseFrame(RaiseList)^.ExceptObject;
        PRaiseFrame(RaiseList)^.ExceptObject:=nil
        end
    end
  end else Result:=DefWindowProc(Window,Message,wParam,lParam)
  //end //end Case
end;

var ThreadWindowClass:TWndClass=(style:0;lpfnWndProc:@ThreadWndProc;cbClsExtra:0;cbWndExtra:0;hInstance:0;hIcon:0;hCursor:0;hbrBackground:0;lpszMenuName:nil;lpszClassName:'TThreadWindow');

procedure AddThread;

  function AllocateWindow:HWND;
  var
    TempClass:TWndClass;
    ClassRegistered:Boolean;
  begin
  ThreadWindowClass.hInstance:=HInstance;
  ClassRegistered:=GetClassInfo(HInstance,ThreadWindowClass.lpszClassName,TempClass);
  if not ClassRegistered or (TempClass.lpfnWndProc<>@ThreadWndProc) then
    begin
    if ClassRegistered then Windows.UnregisterClass(ThreadWindowClass.lpszClassName,HInstance);
    Windows.RegisterClass(ThreadWindowClass)
    end;
  Result:=CreateWindow(ThreadWindowClass.lpszClassName,'',0,0,0,0,0,0,0,HInstance,nil)
  end;

begin
if ThreadCount=0 then ThreadWindow:=AllocateWindow;
Inc(ThreadCount)
end;

procedure RemoveThread;
begin
Dec(ThreadCount);
if ThreadCount=0 then DestroyWindow(ThreadWindow)
end;

function ThreadProc(Thread:TSimpleThread):Integer;
var FreeThread:Boolean;
begin
Thread.Execute;
FreeThread:=Thread.FFreeOnTerminate;
Result:=Thread.FReturnValue;
Thread.FFinished:=True;
Thread.DoTerminate;
if FreeThread then Thread.Free;
EndThread(Result)
end;

constructor TSimpleThread.Create(CreateSuspended:Boolean);
var Flags:Integer;
begin
inherited Create;
AddThread;
FSuspended:=CreateSuspended;
Flags:=0;
if CreateSuspended then Flags:=CREATE_SUSPENDED;
FHandle:=BeginThread(nil,0,@ThreadProc,Pointer(Self),Flags,FThreadID)
end;

destructor TSimpleThread.Destroy;
begin
if not FFinished and not Suspended then
  begin
  Terminate;
  WaitFor
  end;
if FHandle<>0 then CloseHandle(FHandle);
inherited Destroy;
RemoveThread
end;

procedure TSimpleThread.CallOnTerminate;
begin
FOnTerminate(Self)
end;

procedure TSimpleThread.DoTerminate;
begin
if Assigned(FOnTerminate) then Synchronize(CallOnTerminate)
end;

const Priorities:array [TThreadPriority] of Integer=(THREAD_PRIORITY_IDLE,THREAD_PRIORITY_LOWEST,THREAD_PRIORITY_BELOW_NORMAL,THREAD_PRIORITY_NORMAL,THREAD_PRIORITY_ABOVE_NORMAL,THREAD_PRIORITY_HIGHEST,THREAD_PRIORITY_TIME_CRITICAL);

function TSimpleThread.GetPriority:TThreadPriority;
var
  P:Integer;
  I:TThreadPriority;
begin
P:=GetThreadPriority(FHandle);
Result:=tpNormal;
for I:=Low(TThreadPriority) to High(TThreadPriority) do if Priorities[I]=P then Result:=I
end;

procedure TSimpleThread.SetPriority(Value:TThreadPriority);
begin
SetThreadPriority(FHandle,Priorities[Value])
end;

procedure TSimpleThread.Synchronize(Method:TThreadMethod);
begin
//if FMainThreadWaiting then raise EThread.CreateRes(SMainThreadWaiting);//fehlt ab Delphi 3 schon wieder, Meldung lautet: Synchronisierung ausgelst als Haupt-VCL-Thread sich in einem WaitFor-Aufruf befand
FSynchronizeException:=nil;
FMethod:=Method;
SendMessage(ThreadWindow,CM_EXECPROC,0,Longint(Self));
if Assigned(FSynchronizeException) then raise FSynchronizeException
end;

procedure TSimpleThread.SetSuspended(Value:Boolean);
begin
if Value<>FSuspended then if Value then Suspend else Resume
end;

procedure TSimpleThread.Suspend;
begin
FSuspended:=True;
SuspendThread(FHandle)
end;

procedure TSimpleThread.Resume;
begin
if ResumeThread(FHandle)=1 then FSuspended:=False
end;

procedure TSimpleThread.Terminate;
begin
FTerminated:=True
end;

function TSimpleThread.WaitFor:dword;//Integer;
begin
if GetCurrentThreadID=MainThreadID then FMainThreadWaiting:=True;
WaitForSingleObject(FHandle,INFINITE);
GetExitCodeThread(FHandle,Result)
end;

initialization
MainThreadID:=GetCurrentThreadID

finalization

end.
